/*
 * Copyright © Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 2 or 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <mir/logging/dumb_console_logger.h>

#include <iostream>

namespace ml = mir::logging;

void ml::DumbConsoleLogger::log(Severity severity,
                                const std::string& message,
                                const std::string& component)
{
    std::ostream& out = severity <Severity::informational ? std::cerr : std::cout;

    format_message(out, severity, message, component);
}

mir::logging::DumbConsoleLogger::~DumbConsoleLogger()
{
}
